/*
 * This file is part of Cockpit.
 *
 * Copyright (C) 2015 Red Hat, Inc.
 *
 * Cockpit is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2.1 of the License, or
 * (at your option) any later version.
 *
 * Cockpit is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with Cockpit; If not, see <http://www.gnu.org/licenses/>.
 */

import $ from "jquery";
import mustache from "mustache";

//import cockpit from "cockpit";
import * as plot from "./plot";
import machines from "./machines";
import * as machine_dialogs from "./machine-dialogs";
import * as machine_info from "./machine-info";
import "./patterns";

const _ = cockpit.gettext;

// 保存资产和分组信息的全局变量
var g_asserts_bak = {};
var g_groups_bak = {};

/* Handles an href link to a server */
$(document).on("click", "a[data-address]", function(ev) {
  if (ev.target.localName === "input" && ev.target.className == "dashboard-host-checkbox") {
    return ;
  }
    cockpit.jump("/", $(this).attr("data-address"));
    ev.preventDefault();
    return false;
});

var common_plot_options = {
    legend: { show: false },
    series: { shadowSize: 0 },
    xaxis: { tickColor: "#d1d1d1", mode: "time", tickFormatter: plot.format_date_tick, minTickSize: [ 1, 'minute' ] },
    // The point radius influences the margin around the grid even if
    // no points are plotted.  We don't want any margin, so we set the
    // radius to zero.
    points: { radius: 0 },
    grid: {
        borderWidth: 1,
        borderColor: "#e1e6ed",
        //hoverable: true,
        autoHighlight: false
    }
};

var resource_monitors = [
    { selector: "#dashboard-plot-0",
      plot: function (info) {
          return {
              direct: [
                  "kernel.all.cpu.nice",
                  "kernel.all.cpu.user",
                  "kernel.all.cpu.sys"
              ],
              internal: [
                  "cpu.basic.nice",
                  "cpu.basic.user",
                  "cpu.basic.system"
              ],
              units: "millisec",
              derive: "rate",
              factor: 0.1 / info.cpus // millisec / sec -> percent
          };
      },
      options: { yaxis: { tickColor: "#e1e6ed",
                          tickFormatter: function(v) { return v + "%" } } },
      ymax_min: 100
    },
    { selector: "#dashboard-plot-1",
      plot: {
          direct: [
              "mem.util.used"
          ],
          internal: [
              "memory.used"
          ],
          units: "bytes",
      },
      options: {
          yaxis: {
              ticks: plot.memory_ticks,
              tickColor: "#e1e6ed",
              tickFormatter: plot.format_bytes_tick
          }
      },
      ymax_unit: 100000000
    },
    { selector: "#dashboard-plot-2",
      plot: {
          direct: [
              "network.interface.total.bytes"
          ],
          internal: [
              "network.interface.rx",
              "network.interface.tx"
          ],
          units: "bytes",
          'omit-instances': [ "lo" ],
          derive: "rate"
      },
      options: {
          yaxis: {
              tickColor: "#e1e6ed",
              tickFormatter: plot.format_bits_per_sec_tick
          }
      },
      ymax_min: 100000
    },
    { selector: "#dashboard-plot-3",
      plot: {
          direct: [
              "disk.dev.total_bytes"
          ],
          internal: [
              "block.device.read",
              "block.device.written"
          ],
          units: "bytes",
          derive: "rate"
      },
      options: {
          yaxis: {
              ticks: plot.memory_ticks,
              tickColor: "#e1e6ed",
              tickFormatter: plot.format_bytes_per_sec_tick
          }
      },
      ymax_min: 100000
    }
];

var permission = cockpit.permission({ admin: true });
$(permission).on("changed", update_servers_privileged);

function update_servers_privileged() {
    let element = $(".servers-privileged");
    element.update_privileged(
        permission, cockpit.format(
            _("The user <b>$0</b> is not permitted to manage servers"),
            permission.user ? permission.user.name : '')
    );
}

let g_monitor_vue_pagination = false;
let g_cur_page_monitor = 1;
let g_page_size_monitor = 20;

PageDashboard.prototype = {
    _init: function() {
        this.id = "dashboard";
    },

    getTitle: function() {
        return null;
    },

    setup: function() {
        var self = this;

        self.machines = machines.instance();
        self.infos = { };

        var current_monitor = 0;

        var renderer = host_renderer($("#dashboard-hosts .list-group"));
        $(self.machines).on("added.dashboard", renderer);
        $(self.machines).on("removed.dashboard", renderer);
        $(self.machines).on("updated.dashboard", renderer);

        $('#dashboard .nav-tabs li').click(function () {
            set_monitor(parseInt($(this).data('monitor-id'), 10));
        });

        function set_monitor(id) {
            $('#dashboard .nav-tabs li').removeClass("active");
            $('#dashboard .nav-tabs li[data-monitor-id=' + id + ']').addClass("active");
            current_monitor = id;
            $('.dashboard-plot').hide();
            $(resource_monitors[id].selector).show();
            plot_refresh();
        }

        plot_init();
        set_monitor(current_monitor);
        plot.setup_plot_controls($('#dashboard'), $('#dashboard-toolbar'), self.plots);

        function highlight(item, val) {
            //item.toggleClass("highlight-ct", val);
            var ser = series[item.attr("data-address")];
            if (ser) {
                ser.forEach(function (s) {
                    s.options.lines.lineWidth = val ? 3 : 2;
                    if (val)
                        s.move_to_front();
                });
                plot_refresh();
            }
        }

        var series = { };

        function update_series() {
            var refresh = false;

            var seen = { };
            $.each(series, function(addr) {
                seen[addr] = true;
            });

            $("#dashboard-hosts .list-group-item").each(function() {
                var item = $(this);
                var addr = item.attr("data-address");
                var machine = self.machines.lookup(addr);
                var info = self.infos[addr];

                if (!machine || machine.state != "connected")
                    return;

                // 勾选框去掉，不然获取不到资产实时数据
                // 过滤掉没有选中的
                // let checkbox = $(this).children("input.dashboard-host-checkbox");
                // let checked = checkbox.prop("checked");
                // if (!checked) {
                //   return;
                // }
                highlight(item, true);
                
                if (!info) {
                    self.infos[addr] = true;
                    machine_info.cpu_ram_info(machine.connection_string)
                            .done(function (info) {
                                self.infos[addr] = info;
                                update_series();
                            });
                    return;
                } else if (info === true) {
                    // still retrieving
                    return;
                }

                delete seen[addr];
                if (!series[addr]) {
                    series[addr] = plot_add(addr, info);
                }
                series[addr].forEach(function (s) {
                    $(s)
                            .off('hover')
                            .on('hover', function(event, val) {
                            });
                    var color = machine.color;
                    if (s.options.color != color) {
                        refresh = true;
                        s.options.color = color;
                    }
                });
            });

            $.each(seen, function(addr) {
                series[addr].forEach(function (s) { s.remove() });
                delete series[addr];
            });

            if (refresh)
                plot_refresh();
        }

        function host_renderer(target) {
            var template = $("#dashboard-hosts-tmpl").html();
            mustache.parse(template);

            function render_avatar() {
                if (this.state == "failed")
                    return "../shell/images/server-error.png";
                else if (this.avatar)
                    return this.avatar;
                else
                    return "../shell/images/server-small.png";
            }

            function aria_role() {
                if (this.state == "failed")
                    return "alert";
                else
                    return "presentation";
            }

            function alt_text() {
                if (this.state == "failed")
                    return _("Connection Error");
                else
                    return "";
            }

            function avatar_display() {
                if (this.restarting)
                    return "hidden";
                else
                    return "";
            }

            function connecting_display() {
                if (this.restarting)
                    return "";
                else
                    return "hidden";
            }

            function render() {

              function registerEvent() {
                $("#dashboard-hosts-dropdown").change(function(e) {
                  realrender();
                });
                $("#dashboard-hosts-filter").change(function(e){
                  realrender();
                });
                $("input.dashboard-host-checkbox").unbind("click").click(function(e){
                  let assertid = $(this).data("assertid");
                  let checked = $(this).prop("checked");
                  if (assertid == "all") {
                    if (checked) {
                      $("input.dashboard-host-checkbox").prop("checked", true);
                    } else {
                      $("input.dashboard-host-checkbox").prop("checked", false);
                    }
                  }

                  update_series();
                })
              }

              function realrender() {
                let filtertxt = $("#dashboard-hosts-filter").val();
                let filtergrp = $("#dashboard-hosts-dropdown").children("option:selected").val();

                // 根据分组来过滤
                $("#dashboard-hosts-dropdown").html("");
                let html = '<option value="allGrpFlag" selected="">全部</option>';
                $("#dashboard-hosts-dropdown").append($(html));

                for(let i=0;i<g_groups_bak.length;i++) {
                  let html = '<option value="' + g_groups_bak[i].groupid + '">' + g_groups_bak[i].groupname + '</option>';
                  $("#dashboard-hosts-dropdown").append($(html));
                }
                $("#dashboard-hosts-dropdown").val(filtergrp);

                let allAssertsGroup = {};
                for(let i=0;i<g_groups_bak.length;i++) {
                  let groupid = g_groups_bak[i].groupid;
                  let assertsarr = g_groups_bak[i].groupasserts.split(";")
                  for (let j=0;j<assertsarr.length;j++) {
                    let assertid = assertsarr[j];
                    allAssertsGroup[assertid] = groupid;
                  }
                }

                let tmpmachines = [];
                for (let i=0;i<g_asserts_bak.length;i++) {
                  // 过滤搜索字符串
                  let assertid = g_asserts_bak[i].assertid;
                  let assertip = g_asserts_bak[i].assertip;
                  let assertname = g_asserts_bak[i].assertname;
                  if (filtertxt && assertip.indexOf(filtertxt) < 0 && assertname.indexOf(filtertxt) < 0) {
                    continue
                  }
                  // 过滤分组
                  if (filtergrp != "allGrpFlag" && allAssertsGroup[assertid] != filtergrp) {
                    continue
                  }
                  for (let j=0;j<self.machines.list.length;j++) {
                    if (self.machines.list[j].address == assertip) {
                      tmpmachines.push(self.machines.list[j]);
                      tmpmachines[tmpmachines.length-1].assertname = assertname;
                      break;
                    }
                  }
                }

                // 按照 ip 排序
                let sortedtmpmachines = [];
                function cmp_assert(a, b) {
                  if (tmpmachines[a].address == "localhost") return true;
                  else if (tmpmachines[b].address == "localhost") return false;
                  else return tmpmachines[a].address.localeCompare(tmpmachines[b].address)
                }
                Object.keys(tmpmachines).sort(cmp_assert).forEach(function(id){
                  sortedtmpmachines.push(tmpmachines[id])
                })
                tmpmachines = sortedtmpmachines;


                // 过滤分页
                // 分页展示
                if (!g_monitor_vue_pagination) {
                    g_monitor_vue_pagination = new Vue({
                        template: '<el-pagination @size-change="handleSizeChange" @current-change="handleCurrentChange" :current-page="current_page" :page-sizes="[20, 40, 60, 80]" :page-size="page_size" layout="total, sizes, prev, pager, next, jumper" :total="total"> </el-pagination>',
                        el: "#monitor-vue-app-div",
                        methods: {
                            handleSizeChange(val) {
                                g_page_size_monitor = val;
                                console.log("set g_page_size_monitor to:", val)
                                realrender()
                            },
                            handleCurrentChange(val) {
                                g_cur_page_monitor = val;
                                console.log("set g_cur_page_monitor to:", val)
                                realrender()
                            }
                        },
                        data: {
                            current_page: g_cur_page_monitor,
                            page_size: g_page_size_monitor,
                            total: tmpmachines.length
                        }
                    })
                } else {
                    g_monitor_vue_pagination.current_page = g_cur_page_monitor;
                    g_monitor_vue_pagination.page_size = g_page_size_monitor;
                    g_monitor_vue_pagination.total = tmpmachines.length;
                }

                // 只展示这一页的数据
                tmpmachines = tmpmachines.slice( (g_cur_page_monitor-1) * g_page_size_monitor,
                  g_cur_page_monitor * g_page_size_monitor )

                // 修改资产绑定颜色为rgb(a,b,c) => c++ -> b++ -> a++
                var a=0;
                var b=0;
                var c=0;

                for(var i=0;i<tmpmachines.length;i++) {
                    if( i == 254 ){
                        b = b + 1;
                        c = 0;
                    }
                    if( i >= 64770 ){
                        a = a + 1;
                        b = 0;
                        c = 0;
                    }
                    tmpmachines[i].color = "rgb("+a+","+b+","+c+")";
                    c = c + 1;
                }

                var text = mustache.render(template, {
                    machines: tmpmachines,
                    render_avatar: render_avatar,
                    aria_role: aria_role,
                    alt_text: alt_text,
                    avatar_display: avatar_display,
                    connecting_display: connecting_display
                });

                target.html(text);
                update_series();

                // fix bug 32709, add machine color
                $("[data-color]", target).each(function() {
                    $(this).css("border-left-color", $(this).attr("data-color"));
                });

                registerEvent();
              }

              // 根据资产和分组来展示，而不是根据服务器信息来展示，同时支持搜索
              function fetchDataAndRender() {
                let http = cockpit.http({
                        "port": 9091
                    });
                http.get("/getAsserts")
                .fail(function(exception, err) {
                    return console.log("获取资产信息失败，请稍后刷新重试");
                })
                .done(function(res) {
                    // console.log("getAsserts res:", res);

                    let info = $.parseJSON( res );
                    if (! info.succeed) {
                        return console.log("获取资产信息失败，请稍后刷新重试");
                    }
                    g_asserts_bak = info.data;

                    http.get("/getGroups")
                    .fail(function(err){
                        return console.log("获取分组信息失败，请稍后刷新重试");
                    })
                    .done(function(res) {
                        // console.log("get groups res:", res);

                        let info = $.parseJSON( res );
                        if (!info.succeed) {
                            return console.log("获取分组信息失败，请稍后刷新重试");
                        }
                        g_groups_bak = info.data;

                        realrender();
                    })
                })
              }

              fetchDataAndRender();
            }

            /* delay and throttle rendering
               events shouldn't fire continuously anyway,
               so in case of a burst it's better to wait a bit before we start rendering
             */
            function throttled_render() {
                var timer = null;
                return function() {
                    if (timer === null) {
                        timer = window.setTimeout(function () {
                            timer = null;
                            render();
                        }, 500);
                    }
                };
            }
            return throttled_render();
        }

        function plot_refresh() {
            self.plots.forEach(function (p) {
              p.refresh()
            });
        }

        function plot_add(addr, info) {
            var machine = self.machines.lookup(addr);

            if (!machine || machine.state != "connected")
                return null;

            var series = [ ];
            var i = 0;
            resource_monitors.forEach(function (rm) {
                if (self.plots[i]) {
                    var desc = rm.plot;
                    if (rm.plot.apply)
                        desc = rm.plot(info);

                    series.push(self.plots[i].add_metrics_sum_series($.extend({ host: machine.connection_string },
                                                                              desc),
                                                                     { color: machine.color,
                                                                       lines: {
                                                                           lineWidth: 2
                                                                       }
                                                                     }));
                }
                i += 1;
            });
            return series;
        }

        function plot_init() {
            self.plots = [];
            resource_monitors.forEach(function (rm) {
                function setup_hook(flot) {
                    var axes = flot.getAxes();
                    var config = rm;
                    // 获取、处理并发送实时信息到每个对应资产中
                    var datas = flot.getData();
                    var datalist=[];
                    var connectedasslist=[];
                    $.each(datas,function(i,v){
                        connectedasslist.push(v.color);
                        var alltimedatas = datas[i].data;
                        datalist.push(alltimedatas[alltimedatas.length-1][1]);
                    })
                    var data = [{
                        id:"#dashboard-plot-0",
                        datatype:"#dashboard-cpu.dashboard-add-label",
                        factor:function(d){
                            return(Math.round(d)+"%");
                        }
                    },{
                        id:"#dashboard-plot-1",
                        datatype:"#dashboard-mem.dashboard-add-label",
                        factor:function(d){
                            d=d/1073741824; // 1024^3 -> Gib
                            return(parseFloat(d.toFixed(2))+"Gib");
                        }
                    },{
                        id:"#dashboard-plot-2",
                        datatype:"#dashboard-net.dashboard-add-label",
                        factor:function(d){
                            d=d/125000; // byte -> Mbps
                            return(parseFloat(d.toFixed(2))+"Mbps");
                        }
                    },{
                        id:"#dashboard-plot-3",
                        datatype:"#dashboard-io.dashboard-add-label",
                        factor:function(d){
                            d=d/1048576; // B/s -> Mib/s
                            return(parseFloat(d.toFixed(2))+"Mib/s");
                        }
                    }];
                    $("#list-items.list-group-item").each(function(i,v){
                        if(connectedasslist.length == 0){
                            return;
                        }
                        
                        for (var x=0; x<connectedasslist.length; x++){
                            if (connectedasslist[x] == $(this).attr("data-color")){
                                $.each(data,function(indx,vals) {
                                    if (vals.id == rm.selector) {
                                        $(v).find(vals.datatype).text(vals.factor(datalist[x]));
                                    }
                                });
                            }
                        }
                    });
                    
                    if (rm.ymax_unit) {
                        if (axes.yaxis.datamax)
                            axes.yaxis.options.max = Math.ceil(axes.yaxis.datamax / config.ymax_unit) * rm.ymax_unit;
                        else
                            axes.yaxis.options.max = rm.ymax_unit;
                    }

                    if (rm.ymax_min) {
                        if (axes.yaxis.datamax < rm.ymax_min)
                            axes.yaxis.options.max = rm.ymax_min;
                        else
                            axes.yaxis.options.max = null;
                    }

                    axes.yaxis.options.min = 0;
                }

                if (!rm.selector)
                    return;

                var options = $.extend({ setup_hook: setup_hook },
                                       common_plot_options,
                                       rm.options);
                var pl = new plot.Plot($(rm.selector));
                pl.set_options(options);
                self.plots.push(pl);

            });
            series = {};
            update_series();
        }

        $(window).on('resize.dashboard', function () {
            self.plots.forEach(function (p) { p.resize() });
        });

        renderer();
    },

    show: function() {
        update_servers_privileged();
        this.plots[0].resize();
    },

    enter: function() {
    },

    leave: function () {
    }
};

function PageDashboard() {
    this._init();
}

/*
 * INITIALIZATION AND NAVIGATION
 *
 * The code above still uses the legacy 'Page' abstraction for both
 * pages and dialogs, and expects page.setup, page.enter, page.show,
 * and page.leave to be called at the right times.
 *
 * We cater to this with a little compatibility shim consisting of
 * 'dialog_setup', 'dashboard_page_show', and 'page_hide'.
 */

function dashboard_page_show(p, arg) {
    if (p._entered_)
        p.leave();
    p.enter(arg);
    p._entered_ = true;
    $('#' + p.id).show();
    p.show();
}

function dashboard_init() {
    var dashboard_page;

    function navigate() {
        var path = cockpit.location.path;

        if (path.length === 0) {
            dashboard_page_show(dashboard_page);
        } else { /* redirect */
            console.warn("not a dashboard location: " + path);
            cockpit.location = '';
        }

        $("body").removeAttr("hidden");
    }

    cockpit.translate();

    dashboard_page = new PageDashboard();
    dashboard_page.setup();

    $(cockpit).on("locationchanged", navigate);
    navigate();
}

if (window.location.pathname.indexOf("manager/monitor.html") > 0) {
    $(dashboard_init);
}
